@@ -1,5 +1,5 @@
 package Ryu::Source;
-$Ryu::Source::VERSION = '0.010';
+$Ryu::Source::VERSION = '0.011';
 use strict;
 use warnings;
 
@@ -9,6 +9,13 @@ use parent qw(Ryu::Node);
 
 Ryu::Source - base representation for a source of events
 
+=head1 SYNOPSIS
+
+ my $src = Ryu::Source->new;
+ my $chained = $src->map(sub { $_ * $_ })->prefix('value: ')->say;
+ $src->emit($_) for 1..5;
+ $src->finish;
+
 =head1 DESCRIPTION
 
 This is probably the module you'd want to start with, if you were going to be
@@ -17,12 +24,18 @@ point.
 
 =cut
 
+no indirect;
+
 use Future;
-use Syntax::Keyword::Try;
 use curry::weak;
 
 use Log::Any qw($log);
 
+# Implementation note: it's likely that many new methods will be added to this
+# class over time. Most methods have an attempt at "scope-local imports" using
+# namespace::clean functionality, this is partly to make it easier to copy/paste
+# the code elsewhere for testing, and partly to avoid namespace pollution.
+
 =head1 GLOBALS
 
 =head2 $FUTURE_FACTORY
@@ -72,6 +85,7 @@ our %ENCODER = (
         }
     },
 );
+$ENCODER{'UTF-8'} = $ENCODER{utf8};
 
 our %DECODER = (
     utf8 => sub {
@@ -93,10 +107,11 @@ our %DECODER = (
     base64 => sub {
         require MIME::Base64;
         sub {
-            MIME::Base64::decode_base64($_, '');
+            MIME::Base64::decode_base64($_);
         }
     },
 );
+$DECODER{'UTF-8'} = $DECODER{utf8};
 
 =head1 METHODS
 
@@ -425,6 +440,75 @@ sub by_line : method {
     }, $src);
 }
 
+sub prefix {
+    my ($self, $txt) = @_;
+    my $src = $self->chained(label => (caller 0)[3] =~ /::([^:]+)$/);
+    $self->completed->on_ready(sub {
+        shift->on_ready($src->completed) unless $src->completed->is_ready
+    });
+    $self->each_while_source(sub {
+        $src->emit($txt . $_)
+    }, $src);
+}
+
+sub suffix {
+    my ($self, $txt) = @_;
+    my $src = $self->chained(label => (caller 0)[3] =~ /::([^:]+)$/);
+    $self->completed->on_ready(sub {
+        shift->on_ready($src->completed) unless $src->completed->is_ready
+    });
+    $self->each_while_source(sub {
+        $src->emit($_ . $txt)
+    }, $src);
+}
+
+=head2 as_list
+
+Resolves to a list consisting of all items emitted by this source.
+
+=cut
+
+sub as_list {
+    my ($self) = @_;
+    my @data;
+    $self->each(sub {
+        push @data, $_
+    });
+    $self->completed->transform(done => sub { @data })
+}
+
+=head2 as_arrayref
+
+Resolves to a single arrayref consisting of all items emitted by this source.
+
+=cut
+
+sub as_arrayref {
+    my ($self) = @_;
+    my @data;
+    $self->each(sub {
+        push @data, $_
+    });
+    $self->completed->transform(done => sub { \@data })
+}
+
+=head2 as_string
+
+Concatenates all items into a single string.
+
+Returns a L<Future> which will resolve on completion.
+
+=cut
+
+sub as_string {
+    my ($self) = @_;
+    my $data = '';
+    $self->each(sub {
+        $data .= $_;
+    });
+    $self->completed->transform(done => sub { $data })
+}
+
 =head2 combine_latest
 
 =cut
@@ -443,12 +527,11 @@ sub combine_latest : method {
     my %seen;
     for my $idx (0..$#sources) {
         my $src = $sources[$idx];
-        $src->each(sub {
-            return if $combined->completed->is_ready;
+        $src->each_while_source(sub {
             $value[$idx] = $_;
             $seen{$idx} ||= 1;
             $combined->emit([ $code->(@value) ]) if @sources == keys %seen;
-        });
+        }, $combined);
     }
     retain_future(
         Future->needs_any(
@@ -948,6 +1031,8 @@ sub filter {
 =cut
 
 sub emit {
+    use Syntax::Keyword::Try;
+    use namespace::clean qw(try catch finally);
     my $self = shift;
     my $completion = $self->completed;
     for (@_) {
@@ -1201,6 +1286,13 @@ sub chained {
     }
 }
 
+=head2 each_while_source
+
+Like L</each>, but removes the source from the callback list once the
+parent completes.
+
+=cut
+
 sub each_while_source {
     use Scalar::Util qw(refaddr);
     use List::UtilsBy qw(extract_by);
@@ -1218,10 +1310,30 @@ sub DESTROY {
     my ($self) = @_;
     return if ${^GLOBAL_PHASE} eq 'DESTRUCT';
     $log->tracef("Destruction for %s", $self->describe);
-    # warn "destroy for " . $self->label;
     $self->completed->cancel unless $self->completed->is_ready;
 }
 
+sub catch {
+    use Scalar::Util qw(blessed);
+    use namespace::clean qw(blessed);
+    my ($self, $code) = @_;
+    my $src = $self->chained(label => (caller 0)[3] =~ /::([^:]+)$/);
+    $self->completed->on_fail(sub {
+        my @failure = @_;
+        my $sub = $code->(@failure);
+        if(blessed $sub && $sub->isa('Ryu::Source')) {
+            $sub->each_while_source(sub {
+                $src->emit($_)
+            }, $src);
+        } else {
+            $sub->fail(@failure);
+        }
+    });
+    $self->each_while_source(sub {
+        $src->emit($_)
+    }, $src);
+}
+
 1;
 
 __END__
